#include <windows.h>
#include <mmsystem.h>
#include "pixelfun/pixelfun.h"
#include "math.h"
#include "Vector.h"

DWORD* framebuffer;

struct Ray { Vector origin; Vector dir; };
struct Edge { Vector v1; Vector v2; int other_sector; };
struct Sector { int num_edges; Edge* edges; float floor_height, ceiling_height; int wall_texture, floor_texture, ceiling_texture; };

Vector* verts;
Sector* sectors; Sector outside_sector; int cur_sector;
Vector camera_pos; Vector camera_forward; Vector camera_right;
const float aspect_ratio=320.0f/240.0f;
PNGImage* textures;
struct PlaneColumn { int id; bool ready; int invdz[240]; };
PlaneColumn plane_columns[16];
float plane_ray_x[320];
float plane_ray_z[240];
int clip_top, clip_bottom;

void precalc_plane_column(int id, float plane_z)
{
	for(int i=0;i<240;i++)
		plane_columns[id].invdz[i]=(plane_ray_z[i]==0.0f)?0:(int)((camera_pos.z-plane_z)/plane_ray_z[i]*0x100);
	plane_columns[id].ready=true;
}

void draw_column(int screen_x, float world_z, float distance, bool upper, DWORD color, int mode, float tx, int plane_id)
{
	float scale=aspect_ratio/distance;
	int screen_y=120-(int)((world_z-camera_pos.z)*scale*120);
	int start=upper?0:screen_y;
	int end=upper?screen_y:240;
	if(mode==2) // floor/ceiling texture
	{
		if(start<clip_top) start=clip_top;
		if(end>clip_bottom) end=clip_bottom;
		if(upper && end>clip_top) clip_top=end;
		if(!upper && start<clip_bottom) clip_bottom=start;
		DWORD* src=textures[color].data;
		DWORD* dst=framebuffer+start*320+screen_x;
		if(!plane_columns[plane_id].ready)
			precalc_plane_column(plane_id, world_z);
		float dx=camera_forward.x+camera_right.x*plane_ray_x[screen_x];
		float dy=camera_forward.y+camera_right.y*plane_ray_x[screen_x];
		const int texture_scale=32;
		int idx=(int)(dx*(0x100*texture_scale));
		int idy=(int)(dy*(0x100*texture_scale));
		int itexorg_x=(int)(camera_pos.x*(0x100*texture_scale));
		int itexorg_y=(int)(camera_pos.y*(0x100*texture_scale));
		static int invdz, itexpos_x, itexpos_y, tx, ty;
		int* invdzp=plane_columns[plane_id].invdz+start;
		for(int i=start;i<end;i++,dst+=320)
		{
			invdz=*invdzp++;
			itexpos_x=itexorg_x+((idx*invdz)>>8);
			itexpos_y=itexorg_y+((idy*invdz)>>8);
			tx=itexpos_x>>8;
			ty=itexpos_y>>8;
			*dst=src[((ty&63)<<6)+(tx&63)];
		}
	}
	else if(mode==1) // wall texture
	{
		DWORD* src=textures[color].data+(((int)tx)&63)*64;
		float dy=16.0f/120/scale;
		float ty=0.0f;
		if(upper) ty=-screen_y*dy;
		if(start<clip_top)
		{
			ty+=(clip_top-start)*dy;
			start=clip_top;
		}
		if(end>clip_bottom) end=clip_bottom;
		if(upper && end>clip_top) clip_top=end;
		if(!upper && start<clip_bottom) clip_bottom=start;
		int idy=(int)(dy*0x10000);
		int ity=(int)(ty*0x10000);
		DWORD* dst=framebuffer+start*320+screen_x;
		for(int i=start;i<end;i++,dst+=320,ity+=idy)
			*dst=*(src+((ity>>16)&63));
	}
}

bool ray_edge_intersect(Ray& ray, Edge& edge, Vector& hitpoint, float& hitx) // if ray and edge intersect, set hitpoint to the intersection point
{
	Vector edge_tangent=edge.v2 - edge.v1;
	float edge_length=(edge.v2 - edge.v1).Length();
	edge_tangent*=1.0f/edge_length;
	Vector edge_normal=Vector(-edge_tangent.y, edge_tangent.x, 0.0f);
	float dy=-ray.dir*edge_normal;
	if(dy<0.0f) // backfacing
		return false;
	float dx=ray.dir*edge_tangent;
	float xstart=(ray.origin - edge.v1)*edge_tangent;
	float ydist=(ray.origin - edge.v1)*edge_normal;
	if(ydist<0.0f) // edge behind ray
		return false;
	hitx=xstart + dx/dy*ydist; // position of hitpoint along edge segment
	if(hitx<0.0f || hitx>edge_length) // hit outside line segment
		return false;
	hitpoint=edge.v1 + edge_tangent*hitx;
	return true;
}

void render_column(int x, int sector_id, Ray ray)
{
	Sector* this_sector=&sectors[sector_id];
	int closest_hit=-1;
	float closest_dist=1000000.0f;
	Vector closest_hitpoint;
	float closest_hitx;
	for(int i=0;i<this_sector->num_edges;i++) // find closest edge intersection (can be multiple if room is concave)
	{
		Vector hitpoint;
		float hitx;
		if(ray_edge_intersect(ray, this_sector->edges[i], hitpoint, hitx))
		{
			float dist=(hitpoint - camera_pos)*camera_forward;
			if(dist<closest_dist)
			{
				closest_dist=dist;
				closest_hitpoint=hitpoint;
				closest_hitx=hitx;
				closest_hit=i;
			}
		}
	}
	if(closest_hit==-1) return; // no hit at all, which means we're not in the sector we though we were (or the level is broken)
	Sector* other_sector=&outside_sector; // default if the edge we hit doesn't connect to anything (we'll see the walls of the infinite outside sector)
	Edge* edge=&this_sector->edges[closest_hit];
	if(edge->other_sector!=-1) // the hit edge is connected to another sector
		other_sector=&sectors[edge->other_sector]; // we should now render walls of the other sector instead of the outside one
	draw_column(x, this_sector->floor_height,	closest_dist, false, this_sector->floor_texture, 2, 0.0f, sector_id*2+0);	// floor
	draw_column(x, this_sector->ceiling_height, closest_dist, true,	this_sector->ceiling_texture, 2, 0.0f, sector_id*2+1);// ceiling
	draw_column(x, other_sector->floor_height,	closest_dist, false,other_sector->wall_texture, 1, closest_hitx*16.0f, 0);	// lower wall
	draw_column(x, other_sector->ceiling_height,closest_dist, true,	other_sector->wall_texture, 1, closest_hitx*16.0f, 0);	// upper wall
	if(clip_top!=clip_bottom && edge->other_sector!=-1) // still screen space left, and the hit edge is connected to another sector -> recurse into it
	{
		ray.origin=closest_hitpoint; // move ray so that it starts from the hitpoint
		render_column(x, edge->other_sector, ray); // go through this edge, render back-to-front
	}
}

void build_level(); void destroy_level(); void timing(int frames);

void main()
{
	framebuffer=OpenWindow("raycamber", 320, 240);
	build_level();
	cur_sector=0;
	camera_pos=Vector(-5.0f, 0.0f, 2.0f);
	float t=5.0f;
	for(int i=0;i<240;i++) plane_ray_z[i]=(float)(i-120)/120/aspect_ratio;
	for(int i=0;i<320;i++) plane_ray_x[i]=(float)(i-160)/160;
//	timing(0);
	for(;;)//int fi=0;fi<1000;fi++)
	{
		if(WindowClosed()) break;
		t+=0.01f;
		float camera_angle=sin(t*1.2f)*0.7f; // rotate left and right
		camera_pos.z=1.5f+sin(t*2.5f)*0.5f; // bob up and down
		camera_pos.x=-2.0f+sin(t*0.7f)*3.0f; // move back and forth
		camera_pos.y=0.0f+cos(t*0.4f)*3.0f; // move back and forth
		camera_forward=Vector(cos(camera_angle), sin(camera_angle), 0.0f);
		camera_right=Vector(camera_forward.y, -camera_forward.x, 0.0f);
		sectors[1].floor_height=-0.5f+fabs(sin(t*1.9f)*2.0f);
		sectors[1].ceiling_height=2.0f+fabs(sin(t*1.9f)*2.0f);
		Ray ray;
		for(int i=0;i<16;i++) // clear precalc arrays
			plane_columns[i].ready=false;
		DWORD* dst=framebuffer;
		for(int x=0;x<320;x++) // render all pixel columns
		{
			ray.origin=camera_pos;
			ray.dir=camera_forward+camera_right*plane_ray_x[x];
			clip_top=0;
			clip_bottom=240;
			render_column(x, cur_sector, ray);
		}
		UpdateWindow();
	}
//	timing(1000);
	destroy_level();
	HKFree();
	CloseWindow();
}

// -------------------------------------------------------

void timing(int frames)
{
	static DWORD t0=0;
	if(frames==0)
		t0=timeGetTime();
	else
	{
		DWORD t1=timeGetTime();
		FILE* file=fopen("timing.txt", "w");
		int msecs=t1-t0;
		fprintf(file, "%i frames in %i ms = %i fps", frames, msecs, frames*1000/msecs);
		fclose(file);
	}
}

void rotate_texture(int id)
{
	DWORD temp[64*64];
	for(int y=0;y<64;y++)
		for(int x=0;x<64;x++)
			temp[x*64+y]=textures[id].data[y*64+x];
	for(int i=0;i<64*64;i++)
		textures[id].data[i]=temp[i];
}

void build_level()
{
	textures=new PNGImage[7];
	HKPtr(LoadPNG("wall1.png", textures[0]));
	rotate_texture(0);
	HKPtr(LoadPNG("wall2.png", textures[1]));
	rotate_texture(1);
	HKPtr(LoadPNG("wall3.png", textures[2]));
	rotate_texture(2);
	HKPtr(LoadPNG("floor1.png", textures[3]));
	HKPtr(LoadPNG("floor2.png", textures[4]));
	HKPtr(LoadPNG("floor3.png", textures[5]));
	HKPtr(LoadPNG("floor4.png", textures[6]));

	// default sector used for all areas outside actual map sectors (only its walls will be seen)
	outside_sector.floor_height=0.0f;
	outside_sector.ceiling_height=0.0f;
//	outside_sector.wall_color=0x776666;
	outside_sector.wall_texture=0;

	// global array of vertices
	verts=new Vector[7];
	for(int i=0;i<7;i++)
		verts[i].z=0.0f;
	// the main room
	verts[0].x=-10.0f; verts[0].y= 5.0f;
	verts[1].x=-10.0f; verts[1].y=-5.0f;
	verts[2].x= 10.0f; verts[2].y=-5.0f;
	verts[3].x= 10.0f; verts[3].y= 5.0f;
	// the triangular section
	verts[4].x= 8.0f; verts[4].y= 0.0f;
	verts[5].x= 2.0f; verts[5].y= 4.0f;
	verts[6].x= 2.0f; verts[6].y=-4.0f;

	sectors=new Sector[2];

	sectors[0].num_edges=7;
	sectors[0].edges=new Edge[7];
	sectors[0].edges[0].v1=verts[0];
	sectors[0].edges[0].v2=verts[1];
	sectors[0].edges[0].other_sector=-1;
	sectors[0].edges[1].v1=verts[1];
	sectors[0].edges[1].v2=verts[2];
	sectors[0].edges[1].other_sector=-1;
	sectors[0].edges[2].v1=verts[2];
	sectors[0].edges[2].v2=verts[3];
	sectors[0].edges[2].other_sector=-1;
	sectors[0].edges[3].v1=verts[3];
	sectors[0].edges[3].v2=verts[0];
	sectors[0].edges[3].other_sector=-1;
	sectors[0].edges[4].v1=verts[5];
	sectors[0].edges[4].v2=verts[4];
	sectors[0].edges[4].other_sector=1;
	sectors[0].edges[5].v1=verts[6];
	sectors[0].edges[5].v2=verts[5];
	sectors[0].edges[5].other_sector=1;
	sectors[0].edges[6].v1=verts[4];
	sectors[0].edges[6].v2=verts[6];
	sectors[0].edges[6].other_sector=1;
	sectors[0].floor_height=0.0f;
	sectors[0].ceiling_height=3.0f;
	sectors[0].wall_texture=1;
	sectors[0].floor_texture=3;
	sectors[0].ceiling_texture=5;

	sectors[1].num_edges=3;
	sectors[1].edges=new Edge[3];
	sectors[1].edges[0].v1=verts[4];
	sectors[1].edges[0].v2=verts[5];
	sectors[1].edges[0].other_sector=0;
	sectors[1].edges[1].v1=verts[5];
	sectors[1].edges[1].v2=verts[6];
	sectors[1].edges[1].other_sector=0;
	sectors[1].edges[2].v1=verts[6];
	sectors[1].edges[2].v2=verts[4];
	sectors[1].edges[2].other_sector=0;
	sectors[1].floor_height=-0.2f;
	sectors[1].ceiling_height=2.5f;
	sectors[1].wall_texture=2;
	sectors[1].floor_texture=4;
	sectors[1].ceiling_texture=6;
}

void destroy_level()
{
	delete[] sectors[0].edges;
	delete[] sectors[1].edges;
	delete[] sectors;
	delete[] verts;
	delete[] textures;
}

